@@ -57,6 +57,7 @@ class Sql14Store:
 
         id = sa.Column(sa.Integer, primary_key=True)
         name = sa.Column(sa.String, unique=True, nullable=False)
+        vsns = orm.relationship("AppVsn", lazy="joined", back_populates="app")
 
         def __repr__(self):
             return f"<Sql14Store.App(id='{self.id}')>"
@@ -65,9 +66,12 @@ class Sql14Store:
         __tablename__ = "app_vsns"
 
         id = sa.Column(sa.Integer, primary_key=True)
-        vsn = sa.Column(sa.String, unique=True, nullable=False)
+        vsn = sa.Column(sa.String, nullable=False)
         app_id = sa.Column(sa.Integer, sa.ForeignKey("apps.id"), nullable=False)
         app = orm.relationship("App", lazy="joined")
+        app = orm.relationship("App", lazy="joined", back_populates="vsns")
+
+        vsn_app_unique = sa.UniqueConstraint('vsn", "app_id')
 
         def __repr__(self):
             return f"<Sql14Store.AppVsn(id='{self.id}')>"
@@ -78,7 +82,7 @@ class Sql14Store:
         id = sa.Column(sa.Integer, primary_key=True)
         when_ts = sa.Column(sa.Float, nullable=False, default=lambda: datetime.now().timestamp())
         process_id = sa.Column(sa.Integer, sa.ForeignKey("processes.id"), nullable=False)
-        digest = sa.Column(sa.BINARY, nullable=True, unique=True)
+        digest = sa.Column(sa.BINARY, nullable=True)
         stacktrace_id = sa.Column(sa.Integer, sa.ForeignKey("stacktraces.id"), nullable=True)
         process = orm.relationship("Process", lazy="joined")
         data = orm.relationship("EventData", back_populates="event", lazy="joined")
@@ -138,7 +142,6 @@ class Sql14Store:
         self._setup_tables()
 
     def _create_engine(self, dsn, *args, echo=False, **kwargs) -> sa.engine.base.Engine:
-        print(f"GO {dsn}")
         return sa.create_engine(dsn, echo=echo)
 
     def _create_session_factory(self, engine: sa.engine.base.Engine, *args, **kwargs) -> orm.session.sessionmaker:
@@ -156,74 +159,112 @@ class Sql14Store:
             App = self.App
             app_record = App(name=name)
             session.add(app_record)
+            session.flush()
+            session.refresh(app_record)
 
             if vsn is not None:
                 AppVsn = self.AppVsn
-                event_record = AppVsn(vsn=vsn, app_id=app_record.id)
-                session.add(event_record)
+                app_vsn_record = AppVsn(vsn=vsn, app_id=app_record.id)
+                session.add(app_vsn_record)
+                session.flush()
+                session.refresh(app_vsn_record)
 
             session.flush()
 
             return app_record
 
+    def insert_app_vsn(self, app_id: int, vsn: str) -> Sql14Store.AppVsn:
+        with self.session() as session:
+            app_vsn_record = self.AppVsn(vsn=vsn, app_id=app_id)
+            session.add(app_vsn_record)
+            session.flush()
+            session.refresh(app_vsn_record)
+
+            return app_vsn_record
+
     def get_environment_by_id(self, id: int) -> Optional[Sql14Store.Environment]:
         with self.session() as session:
-            return session.query(self.Environment).get(id)
+            return self._get_environment_by_id(session, id)
+
+    def _get_environment_by_id(self, session, id: int) -> Optional[Sql14Store.Environment]:
+        return session.query(self.Environment).get(id)
+
+    def get_environment_by_name(self, name: str) -> Optional[Sql14Store.Environment]:
+        with self.session() as session:
+            return self._get_environment_by_name(session, name)
+
+    def _get_environment_by_name(self, session, name: str) -> Optional[Sql14Store.Environment]:
+        env_records = session.query(self.Environment).filter(self.Environment.name == name).all()
+        return None if len(env_records) == 0 else env_records[0]
 
     def get_app_by_id(self, id: int) -> Optional[Sql14Store.App]:
         with self.session() as session:
-            return session.query(self.App).get(id)
+            return self._get_app_by_id(session, id)
+
+    def _get_app_by_id(self, session, id: int) -> Optional[Sql14Store.App]:
+        return session.query(self.App).get(id)
+
+    def get_app_by_name(self, name: str) -> Optional[Sql14Store.App]:
+        with self.session() as session:
+            return self._get_app_by_name(session, name)
+
+    def _get_app_by_name(self, session, name: str) -> Optional[Sql14Store.App]:
+        app_records = session.query(self.App).filter(self.App.name == name).all()
+        return None if len(app_records) == 0 else app_records[0]
 
     def get_app_vsn_by_id(self, id: int) -> Optional[Sql14Store.AppVsn]:
         with self.session() as session:
-            return session.query(self.AppVsn).get(id)
+            app_vsn_record = self._get_app_vsn_by_id(session, id)
+            session.refresh(app_vsn_record)
+            return app_vsn_record
+
+    def _get_app_vsn_by_id(self, session, id: int) -> Optional[Sql14Store.AppVsn]:
+        return session.query(self.AppVsn).get(id)
 
     def get_process_by_id(self, id: int) -> Optional[Sql14Store.Process]:
         with self.session() as session:
-            process_record = session.query(self.Process).get(id)
+            return self._get_process_by_id(session, id)
 
-            if process_record is not None:
-                app_record = session.query(self.App).get(process_record.app_id)
-                app_vsn = None
-                if process_record.app_vsn_id is not None:
-                    app_vsn_record = session.query(self.AppVsn).get(process_record.app_vsn_id)
-                    app_vsn = app_vsn_record.vsn
-                env_name = None
-                if process_record.environment_id is not None:
-                    env_record = session.query(self.Environment).get(process_record.environment_id)
-                    env_name = env_record.name
+    def _get_process_by_id(self, session, id: int) -> Optional[Sql14Store.Process]:
+        process_record = session.query(self.Process).get(id)
 
-            return process_record
+        if process_record is not None:
+            app_record = session.query(self.App).get(process_record.app_id)
+            app_vsn = None
+            if process_record.app_vsn_id is not None:
+                app_vsn_record = session.query(self.AppVsn).get(process_record.app_vsn_id)
+                app_vsn = app_vsn_record.vsn
+            env_name = None
+            if process_record.environment_id is not None:
+                env_record = session.query(self.Environment).get(process_record.environment_id)
+                env_name = env_record.name
+
+        return process_record
 
     def insert_process(self, unstored_process: Process) -> Sql14Store.Process:
         with self.session() as session:
-            if unstored_process.environment is None:
-                environment_id = None
-            else:
-                env_records = (
-                    session.query(self.Environment)
-                    .filter(
-                        self.Environment.name == unstored_process.environment.name,
-                    )
-                    .all()
-                )
-                if len(env_records) == 0:
+            env_record = None
+            environment_id = None
+
+            if unstored_process.environment is not None:
+                env_record = self._get_environment_by_name(session, unstored_process.environment.name)
+
+                if env_record is None:
                     env_record = self.Environment(name=unstored_process.environment.name)
                     session.add(env_record)
                     session.flush()
-                else:
-                    env_record = env_records[0]
+                    session.refresh(env_record)
 
                 environment_id = env_record.id
 
-            app_records = session.query(self.App).filter(self.App.name == unstored_process.app.name).all()
-            if len(app_records) == 0:
+            app_record = self._get_app_by_name(session, unstored_process.app.name)
+
+            if app_record is None:
                 app_record = self.App(name=unstored_process.app.name)
                 session.add(app_record)
-            else:
-                app_record = app_records[0]
 
             app_vsn_id = None
+
             if unstored_process.app_vsn is not None:
                 app_vsn_records = (
                     session.query(self.AppVsn).filter(self.AppVsn.vsn == unstored_process.app_vsn.vsn).all()
@@ -232,6 +273,8 @@ class Sql14Store:
                 if len(app_vsn_records) == 0:
                     app_vsn_record = self.AppVsn(vsn=unstored_process.app_vsn.vsn, app_id=app_record.id)
                     session.add(app_vsn_record)
+                    session.flush()
+                    session.refresh(app_vsn_record)
                     app_vsn_id = app_vsn_record.id
                 else:
                     app_vsn_record = app_vsn_records[0]
@@ -262,7 +305,9 @@ class Sql14Store:
         inc_st: Optional[bool] = None,
         st_len: Optional[int] = None,
         st_data: Optional[List[str, int, str]] = None,
-    ) -> Sql14Store.Event:
+        when_ts: float = None,
+        sync: bool = False,
+    ) -> Optional[Sql14Store.Event]:
         with self.session() as session:
             if inc_st is None:
                 inc_st = Config()["stacktraces"]["include"]["for_non_exceptions" if exc is None else "from_exceptions"]
@@ -285,28 +330,28 @@ class Sql14Store:
                 stacktrace_records = session.query(self.Stacktrace).filter(self.Stacktrace.digest == digest).all()
                 if len(stacktrace_records) == 0:
                     stacktrace_record = self.Stacktrace(digest=digest)
+                    session.add(stacktrace_record)
+                    session.flush()
+
+                    for sf in st.stackframes:
+                        stackframe_record = self.Stackframe(
+                            stacktrace_id=stacktrace_record.id,
+                            filename=sf.filename,
+                            lineno=sf.lineno,
+                            src=sf.src,
+                        )
+                        session.add(stackframe_record)
                 else:
                     stacktrace_record = stacktrace_records[0]
+                    session.add(stacktrace_record)
 
-                session.add(stacktrace_record)
-                session.flush()
                 stacktrace_id = stacktrace_record.id
 
-                session.flush()
-
-                for sf in st.stackframes:
-                    stackframe_record = self.Stackframe(
-                        stacktrace_id=stacktrace_record.id,
-                        filename=sf.filename,
-                        lineno=sf.lineno,
-                        src=sf.src,
-                    )
-                    session.add(stackframe_record)
-
             event_record = self.Event(
                 process_id=process.id,
                 digest=None if data is None else utils.data_digest(data),
                 stacktrace_id=stacktrace_id,
+                when_ts=when_ts,
             )
             session.add(event_record)
 
@@ -320,12 +365,17 @@ class Sql14Store:
             session.commit()
             session.refresh(event_record)
 
-            return event_record
+            return event_record if sync else None
 
-    def get_events_by_data(self, where: Iterable[Tuple[str, str]]) -> List[Sql14Store.Event]:
+    def get_events_by_data(
+        self, where: Iterable[Tuple[str, str]], process_id: Optional[int] = None
+    ) -> List[Sql14Store.Event]:
         with self.session() as session:
             query = session.query(self.Event)
 
+            if process_id is not None:
+                query = query.filter(self.Event.process_id == process_id)
+
             for key, val in where:
                 EventDataAlias = orm.aliased(self.EventData)
                 query = query.join(EventDataAlias).filter(EventDataAlias.key == key).filter(EventDataAlias.val == val)
